/*******************************************************************************
  File : book.c

  Description : Gestion de la bibliothque des ouvertures.
*******************************************************************************/

/*******************************************************************************
  Include files :
*******************************************************************************/

#include "book.h"
#include "debug.h"
#include "hash.h"
#include "moves.h"
#include "tools.h"
#include "uci.h"

/*******************************************************************************
  Constants :
*******************************************************************************/

/* Nom des fichiers utiliss par Sissa : */
#define OPENINGS_FILE        "sissa_book.txt"

/* Valeurs limites pour quelques constantes gnrales : */
#define OPENING_LINES_LIMIT    1024
#define OPENING_LINE_LENGTH     256

/*******************************************************************************
  Data :
*******************************************************************************/

#ifdef DEBUG
int select_move_calls = 0;
int make_book_calls = 0;
#endif /* DEBUG */

/* Fichier contenant les ouvertures : */
FILE *openings_file;

/*******************************************************************************
  External functions :
*******************************************************************************/

/*******************************************************************************
  Function : select_move

  Description : Slection d'un coup dans la bibliothque d'ouvertures.

  Validation level : 1 - book.select_move
*******************************************************************************/
bool select_move (move_s *move) {

/* ------------------------------- Local data ------------------------------- */

bool book_move = false;
bool line_to_keep;
bool move_in_line;

char lan_move[6];
char lan_played_moves[BOOK_PLY_LIMIT][6];
char moves_line[STRING_LENGTH];
char opening_line[STRING_LENGTH];
char opening_lines[OPENING_LINES_LIMIT][BOOK_PLY_LIMIT * 6];
char *pointer;

int i, j;
int count_lines;
int selected_line;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("select_move\n", true, &select_move_calls);
#endif /* DEBUG */

  cpu_start = clock ();

  /* Initialisation,  partir de l'heure actuelle, d'une nouvelle squence
     de nombres pseudo-alatoires pour varier le choix des ouvertures : */
  srand(time(NULL));

  if ((openings_file = fopen ((active_color ? OPENINGS_FILE : \
    OPENINGS_FILE), "r")) == NULL) {

    tools.send_output (true,
      "   WARNING : %s can't open the openings file %s,\n"
      "             however it will continue to play.\n\n",
      PROGRAM_NAME, (active_color ? OPENINGS_FILE : OPENINGS_FILE));
  }
  else {

    /* Premire tape : construction de la liste des coups jous : */

    for (i = 0; i < game_ply; i++) {

      tools.convert_move (played_moves[i], lan_played_moves[i]);
    }


#ifdef DEBUG_HIGH
debug.print("game_ply %d - played moves",game_ply);
while (--i >= 0) {
  debug.print(" .%s.",lan_played_moves[i]);
}
debug.print("\n");
#endif /* DEBUG_HIGH */

    /* Seconde tape : recherche des lignes d'ouvertures qui contiennent la
       squence prcdente : */
    count_lines = 0;
    while (!feof (openings_file) && count_lines < OPENING_LINES_LIMIT) {

      /* Lecture d'une ligne d'ouverture dans le fichier : */
      fgets (opening_line, STRING_LENGTH, openings_file);

      /* Supprime le caractre de retour  la ligne : */
      opening_line[strlen(opening_line) - 1] = '\0';
#ifdef DEBUG_HIGH
debug.print ("\nbook line (size %d) .%s.\n", strlen (opening_line), opening_line);
#endif /* DEBUG_HIGH */

      line_to_keep = true;
      i = game_ply;
      while (--i >= 0 && line_to_keep) {
    
        move_in_line = false;
        strcpy (moves_line, opening_line);
        j = 0;
        for (pointer = strtok (moves_line, " ");
             pointer != NULL && j <= game_ply && !move_in_line;
             pointer = strtok (NULL, " ")) {

#ifdef DEBUG_HIGH
debug.print ("  compare played move .%s. and book move .%s.\n",
             lan_played_moves[i],pointer);
#endif /* DEBUG_HIGH */
          if (!strcmp (lan_played_moves[i], pointer)) {

#ifdef DEBUG_HIGH
debug.print ("  played move .%s. is found in %d\n", pointer, j+1);
#endif /* DEBUG_HIGH */
            move_in_line = true;
          }

          j++;
        }

        if (!move_in_line) {

#ifdef DEBUG_HIGH
debug.print ("  played move .%s. is not found \n", lan_played_moves[i]);
#endif /* DEBUG_HIGH */
          line_to_keep = false;

        }
      }

      if (line_to_keep) {

#ifdef DEBUG_HIGH
debug.print ("  line is correct\n");
#endif /* DEBUG_HIGH */
        strcpy (opening_lines[count_lines++], opening_line);
      }
    }
#ifdef DEBUG_HIGH
debug.print("\n");
#endif /* DEBUG_HIGH */

    if (count_lines) {

      /* Troisime tape : Slection d'une ligne d'ouverture. */
      selected_line = rand() % count_lines;
#ifdef DEBUG_HIGH
debug.print("\n");
#endif /* DEBUG_HIGH */

      if (uci_mode) {

        printf ("info string Openings book line %d/%d\n",
          selected_line + 1, count_lines);
      }
      else {

        tools.send_output (true, "   Openings book line (%d/%d)\n",
          selected_line + 1, count_lines);
      }

#ifdef DEBUG_HIGH
debug.print ("selected_line (%d/%d) .%s.\n\n",
             selected_line+1, count_lines, opening_lines[selected_line]);
#endif /* DEBUG_HIGH */

      /* Recherche du coup non jou dans la ligne slectionne : */
      strcpy (moves_line, opening_lines[selected_line]);
      j = 0;
      for (pointer = strtok (moves_line, " ");
           pointer != NULL && j <= game_ply && !book_move;
           pointer = strtok (NULL, " ")) {

        i = game_ply;
        move_in_line = false;
        while (--i >= 0 && !move_in_line) {
#ifdef DEBUG_HIGH
debug.print ("  compare book move .%s. and played move .%s.\n",
             lan_played_moves[i],pointer);
#endif /* DEBUG_HIGH */

          if (!strcmp (lan_played_moves[i], pointer)) {

            move_in_line = true;
          }
        }

        j++;

        if (!move_in_line) { 

          book_move = true;
          strcpy (lan_move, pointer);
#ifdef DEBUG_HIGH
debug.print ("  move to play .%s. in %d\n", pointer, j);
#endif /* DEBUG_HIGH */
        }
      }

      /* Une dernire vrification que le coup est jouable dans la position
         courante : */
      if (tools.verify_move (lan_move)) {

        tools.convert_lan_move (lan_move, move);

        cpu_end = clock ();

        if (uci_mode) {

          printf ("info string Elapsed time to select book move : %.1f s\n",
            (cpu_end - cpu_start) / (double) CLOCKS_PER_SEC);
        }
        else {
          tools.send_output (true, "   Elapsed time to select book move : %.1f s\n",
           (cpu_end - cpu_start) / (double) CLOCKS_PER_SEC);
        }
      }
      else {

        book_move = false;

        tools.send_output (true,
          "   WARNING : Opening line is probably bad :\n"
          "             %s\n", opening_lines[selected_line]);
      }
    }
    else {

      tools.send_output (true, "\n");
    }

    fclose (openings_file);
  }

  return book_move;
}
/* End of function : select_move */

/*******************************************************************************
  Function : make_book

  Description : Permet de construire le fichier des lignes d'ouvertures  partir
                d'un fichier de parties.

  Validation level : 1 - book.make_book
*******************************************************************************/
void make_book (void) {

/* ------------------------------- Local data ------------------------------- */

/* Fichier contenant la liste des parties : */
FILE *games_file;

char game[STRING_LENGTH];
char lan_move[6];

int book_ply;
int count_lines = 0;
int i, j;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("make_book\n", true, &make_book_calls);
#endif /* DEBUG */

  if (   (games_file = fopen (GAMES_FILE, "r")) == NULL
      || (openings_file = fopen (OPENINGS_FILE, "w")) == NULL) {

    tools.send_output (true,
      "   ERROR : %s can't open the games and openings file\n"
      "           (%s et/ou %s).\n\n", PROGRAM_NAME, GAMES_FILE, OPENINGS_FILE);
  }
  else {

    while (!feof (games_file)) {

      count_lines++;

      /* Les 4 premires lignes correspondent  l'en-tte d'une partie et ne
         sont pas conserves : */
      for (i = 0; i < 5; i++) {
 
        fgets (game, STRING_LENGTH, games_file);
      }

      book_ply = 0;
      i = 0;
      while (i < strlen (game) && book_ply < BOOK_PLY_LIMIT) {

        /* Saut du numro du coup : */
        while ((game[i++]) != (int) '.');

        lan_move[4] = '\0';
        lan_move[5] = '\0';
        /* Lecture du coup des blancs : */
        j = 0;
        while (game[i] != (int) ' ') {
          if (   (game[i] >= '0' && game[i] <= '8')
              || (game[i] >= 'a' && game[i] <= 'r')) {
            lan_move[j++] = game[i];
          }
          i++;
        }

        /* Cas particulier des roques : */
        if (!strncmp (lan_move, "000", 3)) {
          strcpy (lan_move, "e1c1");
        }
        else if (!strncmp (lan_move, "00", 2)) {
          strcpy (lan_move, "e1g1");
        }

        /* Ecriture du coup des blancs : */
        book_ply++;
        fprintf (openings_file, lan_move);
        fprintf (openings_file, " ");

        if (++i < strlen (game)) {

          lan_move[4] = '\0';
          lan_move[5] = '\0';
          /* Lecture du coup des noirs : */
          j = 0;
          while (game[i] != (int) ' ') {
            if (   (game[i] >= '0' && game[i] <= '8')
                || (game[i] >= 'a' && game[i] <= 'r')) {
              lan_move[j++] = game[i];
            }
            i++;
          }

          /* Cas particulier des roques : */
          if (!strncmp (lan_move, "000", 3)) {
            strcpy (lan_move, "e8c8");
          }
          else if (!strncmp (lan_move, "00", 2)) {
            strcpy (lan_move, "e8g8");
          }

          /* Ecriture du coup des noirs : */
          book_ply++;
          fprintf (openings_file, lan_move);
          fprintf (openings_file, " ");
        }
      }
      fprintf (openings_file, "\n");
    }

    fclose (games_file);
    fclose (openings_file);

    tools.send_output (true, "   %s write %d lines in its openings book.\n\n",
      PROGRAM_NAME, count_lines, OPENINGS_FILE);
  }
}
/* End of function : make_book */
/*******************************************************************************
  Module declaration :
*******************************************************************************/
book_module book = {

  select_move,
  make_book
};
