/*******************************************************************************
  File : data.c

  Description : Donnes gnrales du programme avec leur description.
*******************************************************************************/

/*******************************************************************************
  Include files :
*******************************************************************************/

#include "data.h"

/*******************************************************************************
  Data :
*******************************************************************************/

#ifdef DEBUG_EVAL
/* Fichier pour enregistrer les informations de dbogage du module
   d'valuation : */
FILE *debug_eval_file;
#endif /* DEBUG_EVAL */

#ifdef DEBUG_HIGH
/* Fichier pour enregistrer les informations de dbogage : */
FILE *debug_file;
#endif /* DEBUG_HIGH */

#ifdef DEBUG_TREE
/* Fichier pour enregistrer les informations de dbogage du module
   de recherche : */
FILE *debug_tree_file;
#endif /* DEBUG_TREE */

/* Fichier pour enregistrer les informations du programme : */
FILE *dump_file;

/* Fichier journal pour enregistrer les changes avec le programme : */
FILE *log_file;

/* Pour la description d'une position de l'chiquier :
   8 colonnes x 8 ranges (l'indice 0 n'est pas utilis)
  (voir la dfinition de la structure board_s) */
board_s board[9][9];

/* Indique si le programme joue automatiquement  la rception d'un coup :
   (In UCI mode, the engine will always be in forced mode which means
   it should never start calculating or pondering without receiving a
   "go" command first) : */
bool forced_mode;

/* Indique si l'chiquier doit tre affich en mode console : */
bool show_board;

/* Indique si les changes entre le programme et l'xtrieur se font en
   respectant le protocole UCI : */
bool uci_mode;

/* Indique si la bibliothque des ouvertures doit tre utilise : */
bool use_book;

/* Permettent de mesurer le temps de rflexion du programme : */
clock_t cpu_start;
clock_t cpu_end;

/* La position de dpart au format EPD : */
const char classic_start_position[STRING_LENGTH] =
#ifdef DEBUG_POSITION
  { "" };
#else
  { "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1" };
#endif /* DEBUG_POSITION */

/* Masques pour isoler les donnes de la variable castle_status, afin de savoir
   si les roques noir et blanc sont effectus :
  (voir la description de castle_status) */
const int castle_done_mask[2] = { 0x20, 0x10 };

/* Masques pour isoler petit et grand roque de chaque couleur :
  (voir la description de castle_status) */
const int castle_by_color_mask[2] = { 0x3C, 0x33 };

/* Masques pour isoler chaque donne de la variable castle_status :
  (voir la description de castle_status) */
const int castle_data_mask[7] = { 0, 0x1, 0x2, 0x4, 0x8, 0x10, 0x20 };

/* Valeur de base des pices du jeu d'checs : */
const int pieces_value[7] = { 0, 100, 300, 300, 500, 900, 0 };

/* Array to reverse rank : */
const int reverse_rank[2][9] = {
  { 0, 8, 7, 6, 5, 4, 3, 2, 1},
  { 0, 1, 2, 3, 4, 5, 6, 7, 8}
};

/* Couleur du camp qui a le trait : */
int active_color;

/* Temps de base allou pour un coup, calcul au dbut de la partie : */
int base_time;

/* Indique les possibilits de roque blanc et noir :
   - bit 1 petit roque blanc possible
   - bit 2 grand roque blanc possible
   - bit 3 petit roque noir possible
   - bit 4 grand roque noir possible
   Indique que le roi blanc a roqu :
   - bit  9 roque blanc effectu
   - bit 10 roque noir effectu

   Exemple : 0xF -> tous les roques sont possibles
             0x6 -> seuls le petit roque noir et le grand roque blanc sont
                    possibles
*/
int castle_status[MOVES_LIMIT * 2];

/* La couleur joue par le programme : */
int computer_color;

/* Profondeur de recherche atteinte (en demi-coups) : */
int depth;

/* This number is the fullmove number. It will have the value "1" for the first
   move of a game for both White and Black. It is incremented by one immediately
   after each move by Black : */
int fullmove;

/* This number is the count of halfmoves since the begin of the game. */
int game_ply;

/* This number is the count of halfmoves (or ply) since the last pawn advance
   or capturing move. It is used for the fifty move draw rule : */
int halfmove;
int halfmove_history[MOVES_LIMIT * 2];

/* Nombre de noeuds de l'arbre de recherche parcourus par le programme : */
int nodes;

/* Nombre de pices sur l'chiquier, blanches et noires, autres que pions et
   rois. Permet de connatre l'volution de ce nombre pendant la recherche : */
int piece_count;

/* Le nombre de pices de chaque couleur : */
int pieces_number[2];

/* Longueur de la variante principale  chaque demi-coup de l'arbre de
   recherche : */
int pv_length[PV_LIMIT];

/* Nombre de noeuds parcourus lors de la phase de recherche des situations
   calmes pour valuer la position : */
int qnodes;

/* Rsultat de la partie : */
int result;

/* This number is the count of halfmoves since the start of the search. */
int search_ply;

/* Nombre de pices sur l'chiquier (cf piece_count), au dbut de chaque
   nouvelle recherche du programme : */
int start_piece_count;

/* Temps allou pour la rflexion du programme (en millimes de seconde) : */
int time_for_move;

/* Coup sans signification : */
move_s null_move;

/* Liste des coups jous (en nombre de demi-coups) : */
move_s played_moves[MOVES_LIMIT * 2];

/* Liste des coups de la variante principale  chaque demi-coup de l'arbre de
   recherche : */
move_s pv[PV_LIMIT][PV_LIMIT];

/* Table des pices (16 pices au maximum pour chaque couleur) : */
piece_s pieces[2][17];

/* La case de prise en passant; elle est renseigne seulement si le coup
   prcdent correspond  l'avance d'un pion de deux cases : */
square_s ep_square[MOVES_LIMIT * 2];

/* Position des rois blancs et noirs : */
square_s king[2];

/* Instant de dbut de la rflexion (en millimes de seconde) : */
time_t start_time;
