/*******************************************************************************
  File : data.h

  Description : Fichier en-tte pour les donnes gnrales.
*******************************************************************************/

#ifndef DATA_H
#define DATA_H

/*******************************************************************************
  Include files :
*******************************************************************************/

#include <limits.h>
#include <signal.h>
#include <stdarg.h>
//#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define bool int
#define true 1
#define false 0
/*******************************************************************************
  Constants :
*******************************************************************************/

/* Informations sur le programme : */
#define PROGRAM_AUTHOR       "Christophe J. Mandin"
#define PROGRAM_NAME         "Sissa"
#define PROGRAM_VERSION      "0.12.0 JA"
#define PROGRAM_YEAR         "2009"

/* Ces constantes permettent de choisir diffrents dbogages possibles : */
#define NO_DEBUG
#define NO_DEBUG_EVAL
#define NO_DEBUG_HIGH
#define NO_DEBUG_POSITION
#define NO_DEBUG_TREE

/* Nom des fichiers utiliss par le programme : */
#define DEBUG_EVAL_FILE      "debug_eval.txt"
#define DEBUG_FILE           "debug.txt"
#define DEBUG_TREE_FILE      "debug_tree.txt"
#define DUMP_FILE            "dump.txt"
#define GAMES_FILE           "games.txt"
#define LOG_FILE             "logfile.txt"

/* Valeurs limites pour quelques constantes gnrales : */
#define BOOK_PLY_LIMIT           20
#define DEBUG_NODES_LIMIT     10000
#define DEBUG_NODES_MAX      400000
#define DEBUG_NODES_MIN           0
#define DEBUG_TREE_LIMIT    2000000
#define MIN_MAX_LIMIT       1000000
#define MOVES_LIMIT             256
#define PV_LIMIT               1024
#define STRING_LENGTH          4096

/* Noms des huit colonnes de l'chiquier : */
#define A 1
#define B 2
#define C 3
#define D 4
#define E 5
#define F 6
#define G 7
#define H 8

/* Noms des couleurs des deux camps : */
#define BLACK   0
#define WHITE   1

/* Noms pour quelques coups spciaux : */
#define KING_CASTLE    1
#define QUEEN_CASTLE   2
#define EN_PASSANT     3
#define PROMOTION      4

/* Noms des diffrentes phases de la partie : */
#define OPENING      0
#define MIDDLE       1
#define ENDGAME      2
#define UNDEFINED_GS 3

/* Noms des rsultats possibles pour une partie : */
#define NO_RESULT          0
#define BLACK_IS_MATED     1
#define WHITE_IS_MATED     2
#define DRAW_FIFTY_MOVES   3
#define DRAW_REPETITION    4
#define DRAW_STALEMATE     5

/* Noms des diffrentes gnrations de coups possibles : */
#define LEGAL_MOVES     0
#define CAPTURES_ONLY   1
#define PSEUDO_MOVES    2

/*******************************************************************************
  Macro-instructions :
*******************************************************************************/

#define CASTLE_STATUS castle_status[game_ply]
#define EP_SQUARE     ep_square[game_ply]

/* Calcule la distance entre deux cases : */
#define DISTANCE(a,b,c,d) \
  ((a<c?c-a:a-c)>(b<d?d-b:b-d)?(a<c?c-a:a-c):(b<d)?d-b:b-d)

/* Teste si une case est en dehors de l'chiquier : */
#define IS_INSIDE(x, y) \
  ((x > 0 && x < 9 && y > 0 && y < 9) ? true : false)

/* Pour une direction, dtermine la direction oppose : */
#define OPPOSED(d) \
  (d == 0 ? 0 : d > 8 ? (d + 3) % 18 + 1 : (d + 3) % 8 + 1)

/* Retourne le nom de la pice : */
#define PIECE(p) \
  (p == PAWN ? 'p' : p == BISHOP ? 'b' : p == KNIGHT ? 'n' : p == ROOK ? 'r' :\
   p == QUEEN ? 'q' : 'k')

#define RR(c) (reverse_rank[color][c])

/*******************************************************************************
  Types :
*******************************************************************************/

/* Liste des pices de l'chiquier : */
typedef enum { NONE, PAWN, KNIGHT, BISHOP, ROOK, QUEEN, KING } pieces_e;

/* Informations qui dcrivent une case de l'chiquier : */
typedef struct {

  /* Nom de la pice (NONE or from PAWN to KING) : */
  int piece;

  /* Couleur d'une pice (BLACK or WHITE) : */
  int color;

  /* Indice de la fiche associe  la pice : */
  int form;

  /* Indique si la pice a t dplace : */
  int moved;

} board_s;

/* Informations qui dcrivent une pice : */
typedef struct {

  /* Nom de la pice : */
  int piece;

  /* Position sur l'chiquier : */
  int x;
  int y;

  /* Indique si la pice a t capture : */
  bool captured;

  /* Direction (et celle oppose) autorise en cas de clouage : */
  int authorized;

} piece_s;

/* Informations qui permettent d'adresser une case de l'chiquier : */
typedef struct {

  /* Colonne (de A  H) : */
  int x;

  /* Range (de 1  8) : */
  int y;

} square_s;

/* Informations qui dcrivent un coup : */
typedef struct move {

  /* Nom de la pice dplace : */
  int piece;

  /* Couleur de la pice dplace : */
  int color;

  /* Case de dpart de la pice : */
  square_s from;

  /* Case d'arrive de la pice : */
  square_s to;

  /* Nature du coup jou : */
  int status;

  /* Nom de la pice de promotion : */
  int promotion;

  /* Indice de la fiche associe  la pice dplace : */
  int form;

  /* Indice de la fiche associe  la pice capture : */
  int captured_form;

  /* Note pour le tri des coups lors de la phase de recherche : */
  int note;

} move_s;

/*******************************************************************************
  Data :
 (Pour la description des donnes, voir le fichier data.c)
*******************************************************************************/

#ifdef DEBUG_EVAL
extern FILE *debug_eval_file;
#endif /* DEBUG_EVAL */

#ifdef DEBUG_HIGH
extern FILE *debug_file;
#endif /* DEBUG_HIGH */

#ifdef DEBUG_TREE
extern FILE *debug_tree_file;
#endif /* DEBUG_TREE */

extern FILE *dump_file;
extern FILE *log_file;

extern board_s board[9][9];

extern bool forced_mode;
extern bool show_board;
extern bool uci_mode;
extern bool use_book;

extern clock_t cpu_start;
extern clock_t cpu_end;

extern const char classic_start_position[STRING_LENGTH];
extern const int castle_done_mask[2];
extern const int castle_by_color_mask[2];
extern const int castle_data_mask[7];
extern const int pieces_value[7];
extern const int reverse_rank[2][9];

extern int active_color;
extern int base_time;
extern int castle_status[MOVES_LIMIT * 2];
extern int computer_color;
extern int depth;
extern int fullmove;
extern int game_ply;
extern int halfmove;
extern int halfmove_history[MOVES_LIMIT * 2];
extern int nodes;
extern int piece_count;
extern int pieces_number[2];
extern int pv_length[PV_LIMIT];
extern int qnodes;
extern int result;
extern int search_ply;
extern int start_piece_count;
extern int time_for_move;

extern move_s null_move;
extern move_s played_moves[MOVES_LIMIT * 2];
extern move_s pv[PV_LIMIT][PV_LIMIT];

extern piece_s pieces[2][17];

extern square_s ep_square[MOVES_LIMIT * 2];
extern square_s king[2];

extern time_t start_time;

#endif /* DATA_H */
