/*******************************************************************************
  File : debug.c

  Description : Diverses fonctions utilises pour dboguer ou tester le
                programme.
*******************************************************************************/

/*******************************************************************************
  Include files :
*******************************************************************************/

#include "book.h"
#include "data.h"
#include "evaluate.h"
#include "debug.h"
#include "hash.h"
#include "moves.h"
#include "search.h"
#include "sissa.h"
#include "tools.h"
#include "uci.h"

/*******************************************************************************
  Data :
*******************************************************************************/

#ifdef DEBUG
char debug_file_name[STRING_LENGTH];

int debug_file_indice = 0;
int debug_nodes = 1;

int total_calls = 0;
#endif /* DEBUG */

/*******************************************************************************
  External functions :
*******************************************************************************/

/*******************************************************************************
  Function : close_debug_files

  Description : Fermeture de tous les fichiers de debug.

  Validation level : 1 - debug.close_debug_files
*******************************************************************************/
void close_debug_files (void) {

#ifdef DEBUG_EVAL
  fclose (debug_eval_file);
#endif /* DEBUG_EVAL */

#ifdef DEBUG_HIGH
  fclose (debug_file);
#endif /* DEBUG_HIGH */

#ifdef DEBUG_TREE
  fclose (debug_tree_file);
#endif /* DEBUG_TREE */
}
/* End of function : close_debug_files */

/*******************************************************************************
  Function : dump

  Description : Enregistrement dans un fichier des informations mmorises par
                le programme.

  Validation level : 1 - debug.dump
*******************************************************************************/
void dump (void) {

/* ------------------------------- Local data ------------------------------- */

int color;
int i, x, y;

/* ------------------------------ Function code ----------------------------- */

  if ((dump_file = fopen (DUMP_FILE, "w")) == NULL) {

    printf ("   ERROR : %s can't open the dump file %s\n",
      PROGRAM_NAME, DUMP_FILE);
  }

  /* Enregistrement des donnes globales : */
  fprintf (dump_file, "forced_mode        : %d\n", forced_mode);
  fprintf (dump_file, "show_board         : %d\n", show_board);
  fprintf (dump_file, "uci_mode           : %d\n", uci_mode);

  fprintf (dump_file, "active_color       : %d\n", active_color);
  fprintf (dump_file, "computer_color     : %d\n", computer_color);

  fprintf (dump_file, "base_time          : %d\n", base_time);
  fprintf (dump_file, "castle_status[%d]   : %x\n",
           game_ply, castle_status[game_ply]);
  fprintf (dump_file, "depth              : %d\n", depth);
  fprintf (dump_file, "fullmove           : %d\n", fullmove);
  fprintf (dump_file, "game_ply           : %d\n", game_ply);
  fprintf (dump_file, "halfmove           : %d\n", halfmove);
  fprintf (dump_file, "piece_count        : %d\n", piece_count);
  fprintf (dump_file, "pieces_number      : black %2d / white %2d\n",
           pieces_number[BLACK], pieces_number[WHITE]);
  fprintf (dump_file, "search_ply         : %d\n", search_ply);
  fprintf (dump_file, "result             : %d\n", result);
  fprintf (dump_file, "start_piece_count  : %d\n", start_piece_count);
  fprintf (dump_file, "ep_square[%2d]      : %c%d\n", game_ply,
           ep_square[game_ply].x + (int) 'a' - 1, ep_square[game_ply].y);
  fprintf (dump_file, "black king         : %c%d\n",
           king[BLACK].x + (int) 'a' - 1, king[BLACK].y);
  fprintf (dump_file, "white king         : %c%d\n",
           king[WHITE].x + (int) 'a' - 1, king[WHITE].y);

  /* Enregistrement de la table des pices : */
  for (color = BLACK; color <= WHITE; color++) {

    for (i = 1; i <= pieces_number[color]; i++) {

      fprintf (dump_file, "pieces[%d][%2d] piece %d %c%d cap. %d auth. %d\n",
               color, i, pieces[color][i].piece,
               pieces[color][i].x + (int) 'a' - 1, pieces[color][i].y,
               pieces[color][i].captured, pieces[color][i].authorized);
    }
  }

  /* Enregistrement de l'chiquier : */
  for (x = A; x <= H; x++) {

    for (y = 1; y <= 8; y++) {

      fprintf (dump_file, "board[%d][%d] piece %d color %d form %2d moved %d\n",
               x, y, board[x][y].piece, board[x][y].color, board[x][y].form,
               board[x][y].moved);
    }
  }

  fclose (dump_file);
}
/* End of function : dump */

/*******************************************************************************
  Function : new_debug_files

  Description : En raison du volume important des donnes, les fichiers de
                dbogage sont rgulirement dcoups.

  Validation level : 1 - debug.new_debug_files
*******************************************************************************/
void new_debug_files (void) {

#ifdef DEBUG
  if (nodes > DEBUG_NODES_MIN && nodes <= DEBUG_NODES_MAX) {

    if (debug_nodes == DEBUG_NODES_LIMIT) {

      debug_nodes = 1;

      debug.close_debug_files ();
      debug.open_debug_files ();
    }
    else {

      debug_nodes++;
    }
  }
#endif /* DEBUG */
}
/* End of function : new_debug_files */

/*******************************************************************************
  Function : open_debug_files

  Description : Ouverture de tous les fichiers utiliss pour les diffrents
                types de dbogage.

  Validation level : 1 - debug.open_debug_files
*******************************************************************************/
void open_debug_files (void) {

/* ------------------------------- Local data ------------------------------- */

char string_indice[4];

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG_EVAL
  if ((debug_eval_file = fopen (DEBUG_EVAL_FILE, "w")) == NULL) {

    printf ("ERROR : %s can't open the debug file %s\n",
            PROGRAM_NAME, DEBUG_EVAL_FILE);
  }
#endif /* DEBUG_EVAL */

#ifdef DEBUG_HIGH
  debug_file_indice++;

  sprintf (debug_file_name, "debug");
  sprintf (string_indice, "%d", debug_file_indice);
  strcat (debug_file_name, string_indice);
  strcat (debug_file_name, ".txt");

  if (debug_file_indice > 1) {

    printf ("                %8d : %s\n", nodes, debug_file_name);
  }

  if ((debug_file = fopen (debug_file_name, "w")) == NULL) {

    printf ("ERROR : %s can't open the debug file %s\n",
            PROGRAM_NAME, debug_file_name);
  }
#endif /* DEBUG_HIGH */

#ifdef DEBUG_TREE
  if ((debug_tree_file = fopen (DEBUG_TREE_FILE, "w")) == NULL) {

    printf ("ERROR : %s can't open the debug tree file %s\n",
            PROGRAM_NAME, DEBUG_TREE_FILE);
  }
#endif /* DEBUG_TREE */
}
/* End of function : open_debug_files */

/*******************************************************************************
  Function : print

  Description : Pour enregistrer des informations dans le fichier de dbogage.

  Validation level : 1 - debug.print
*******************************************************************************/
int print (const char *format, ...) {

/* ------------------------------- Local data ------------------------------- */

int size = 0;

va_list list;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG_HIGH
  if (nodes >= DEBUG_NODES_MIN && nodes < DEBUG_NODES_MAX) {

    /* Initialisation de la liste des paramtres : */
    va_start (list, format);

    /* Ecriture des donnes dans le fichier de dbogage : */
    size = vfprintf (debug_file, format, list);
    fflush (debug_file);

    /* Fin de lecture de la liste des paramtres : */
    va_end (list);
  }
#endif /* DEBUG_HIGH */

  return size;
}
/* End of function : print */

/*******************************************************************************
  Function : print_informations

  Description : Donne des informations sur l'environnement d'excution du
                programme.

  Validation level : 1 - debug.print_informations
*******************************************************************************/
void print_informations (void) {

  fprintf (log_file, "\nInformations about ANSI C standard types :\n");
  fprintf (log_file, "--------------------------------------------\n");
  fprintf (log_file, "char                   : %d (%d,%d)\n",
           sizeof (char), CHAR_MIN, CHAR_MAX);
  fprintf (log_file, "short                  : %d (%d,%d)\n",
           sizeof (short), SHRT_MIN, SHRT_MAX);
  fprintf (log_file, "int                    : %d (%d,%d)\n",
           sizeof (int), INT_MIN, INT_MAX);
  fprintf (log_file, "long int               : %d (%d,%ld)\n",
           sizeof (long int), LONG_MIN, LONG_MAX);
  fprintf (log_file, "unsigned char          : %d (%d,%d)\n",
           sizeof (unsigned char), 0, UCHAR_MAX);
  fprintf (log_file, "unsigned short         : %d (%d,%d)\n",
           sizeof (unsigned short), 0, USHRT_MAX);
  fprintf (log_file, "unsigned int           : %d (%d,%u)\n",
           sizeof (unsigned int), 0, UINT_MAX);
  fprintf (log_file, "unsigned long int      : %d (%d,%lu)\n",
           sizeof (unsigned long int), 0, ULONG_MAX);
  fprintf (log_file, "float                  : %d\n", sizeof (float));
  fprintf (log_file, "double                 : %d\n", sizeof (double));
  fprintf (log_file, "\n");
}
/* End of function : print_informations */

/*******************************************************************************
  Function : print_statistics

  Description : Donne des informations statistiques sur les appels de fonction.

  Validation level : 1 - debug.print_statistics
*******************************************************************************/
void print_statistics (void) {

#ifdef DEBUG
  fprintf (log_file, "Statistics about function calls :\n");
  fprintf (log_file, "\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "  book module :\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "select_move        : %9d\n", select_move_calls);
  fprintf (log_file, "make_book          : %9d\n", make_book_calls);
  fprintf (log_file, "\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "  evaluate module :\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "check_game_stage   : %9d\n", check_game_stage_calls);
  fprintf (log_file, "evaluate_position  : %9d\n", evaluate_position_calls);
  fprintf (log_file, "\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "  hash module :\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "compute_h_position : %9d\n", compute_h_position_calls);
  fprintf (log_file, "init_h_values      : %9d\n", init_h_values_calls);
  fprintf (log_file, "\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "  moves module :\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "generate           : %9d\n", generate_calls);
  fprintf (log_file, "is_attacked        : %9d\n", is_attacked_calls);
  fprintf (log_file, "make               : %9d\n", make_calls);
  fprintf (log_file, "unmake             : %9d\n", unmake_calls);
  fprintf (log_file, "search_auth_dir    : %9d\n", search_auth_dir_calls);
  fprintf (log_file, "\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "  search module :\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "first_search       : %9d\n", first_search_calls);
  fprintf (log_file, "minmax_search      : %9d\n", minmax_search_calls);
  fprintf (log_file, "pick_move          : %9d\n", pick_move_calls);
  fprintf (log_file, "quiet_search       : %9d\n", quiet_search_calls);
  fprintf (log_file, "sort_moves         : %9d\n", sort_moves_calls);
  fprintf (log_file, "think              : %9d\n", think_calls);
  fprintf (log_file, "\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "  sissa module :\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "computer_to_play   : %9d\n", computer_to_play_calls);
  fprintf (log_file, "init_game          : %9d\n", init_game_calls);
  fprintf (log_file, "main               : %9d\n", main_calls);
  fprintf (log_file, "\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "  tools module :\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "compute_time       : %9d\n", compute_time_calls);
  fprintf (log_file, "convert_lan_move   : %9d\n", convert_lan_move_calls);
  fprintf (log_file, "convert_move       : %9d\n", convert_move_calls);
  fprintf (log_file, "generate_random    : %9d\n", generate_random_calls);
  fprintf (log_file, "is_repeated        : %9d\n", is_repeated_calls);
  fprintf (log_file, "print_board        : %9d\n", print_board_calls);
  fprintf (log_file, "read_input         : %9d\n", read_input_calls);
  fprintf (log_file, "send_output        : %9d\n", send_output_calls);
  fprintf (log_file, "update_pieces      : %9d\n", update_pieces_calls);
  fprintf (log_file, "verify_move        : %9d\n", verify_move_calls);
  fprintf (log_file, "\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "  uci module :\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "go                 : %9d\n", go_calls);
  fprintf (log_file, "info               : %9d\n", info_calls);
  fprintf (log_file, "isready            : %9d\n", isready_calls);
  fprintf (log_file, "position           : %9d\n", position_calls);
  fprintf (log_file, "quit               : %9d\n", quit_calls);
  fprintf (log_file, "uci_cmd            : %9d\n", uci_cmd_calls);
  fprintf (log_file, "ucinewgame         : %9d\n", ucinewgame_calls);
  fprintf (log_file, "\n");
  fprintf (log_file, "****************************************\n");
  fprintf (log_file, "Total calls        : %9d\n", total_calls);
#endif /* DEBUG */
}
/* End of function : print_statistics */

/*******************************************************************************
  Function : test_hash

  Description : Permet de tester les collisions sur les valeurs de hachage.

  Validation level : 1 - book.test_hash
*******************************************************************************/
void test_hash (void) {

/* ------------------------------- Local data ------------------------------- */

#define HASH_TEST_FILE "hash_test.txt"

#define HASH_PLY_LIMIT 120

/* Fichier contenant la liste des parties : */
FILE *games_file;
FILE *hash_test_file;

char game[STRING_LENGTH];
char lan_move[6];

int hash_ply;
int count_games = 0;
int i, j;

move_s move;

/* ------------------------------ Function code ----------------------------- */

  if (   (games_file = fopen (GAMES_FILE, "r")) == NULL
      || (hash_test_file = fopen (HASH_TEST_FILE, "w")) == NULL) {

    tools.send_output (true,
      "   ERROR : %s can't open the games and hash test file\n"
      "           (%s et/ou %s).\n\n", PROGRAM_NAME, GAMES_FILE, 
      HASH_TEST_FILE);
  }
  else {

    while (!feof (games_file) && count_games < 100) {

      count_games++;

      fprintf (hash_test_file, "%4d) ", count_games);

      /* Les 4 premires lignes correspondent  l'en-tte d'une partie et ne
         sont pas conserves : */
      for (i = 0; i < 5; i++) {
 
        fgets (game, STRING_LENGTH, games_file);
      }

      hash_ply = 0;
      i = 0;
      while (i < strlen (game) && hash_ply < HASH_PLY_LIMIT) {

        /* Saut du numro du coup : */
        while ((game[i++]) != (int) '.');

        lan_move[4] = '\0';
        lan_move[5] = '\0';
        /* Lecture du coup des blancs : */
        j = 0;
        while (game[i] != (int) ' ') {
          if (   (game[i] >= '0' && game[i] <= '8')
              || (game[i] >= 'a' && game[i] <= 'r')) {
            lan_move[j++] = game[i];
          }
          i++;
        }

        /* Cas particulier des roques : */
        if (!strncmp (lan_move, "000", 3)) {
          strcpy (lan_move, "e1c1");
        }
        else if (!strncmp (lan_move, "00", 2)) {
          strcpy (lan_move, "e1g1");
        }

        tools.convert_lan_move (lan_move, &move);
        moves.make (move);
        for (j = 0; j < game_ply; j++) {
          if (h_position_history[j] == h_position) {
              printf ("game %4d - game_ply %3d = %8x and game_ply %3d = %8x\n",
                count_games, j, h_position_history[j], game_ply, h_position);
          }
        }

        /* Ecriture du coup des blancs : */
        hash_ply++;
        fprintf (hash_test_file, "%d. ", hash_ply);
        fprintf (hash_test_file, lan_move);
        fprintf (hash_test_file, " %8x ", h_position);

        if (++i < strlen (game)) {

          lan_move[4] = '\0';
          lan_move[5] = '\0';
          /* Lecture du coup des noirs : */
          j = 0;
          while (game[i] != (int) ' ') {
            if (   (game[i] >= '0' && game[i] <= '8')
                || (game[i] >= 'a' && game[i] <= 'r')) {
              lan_move[j++] = game[i];
            }
            i++;
          }

          /* Cas particulier des roques : */
          if (!strncmp (lan_move, "000", 3)) {
            strcpy (lan_move, "e8c8");
          }
          else if (!strncmp (lan_move, "00", 2)) {
            strcpy (lan_move, "e8g8");
          }

          tools.convert_lan_move (lan_move, &move);
          moves.make (move);
          for (j = 0; j < game_ply; j++) {
            if (h_position_history[j] == h_position) {
              printf ("game %4d - game_ply %3d = %8x and game_ply %3d = %8x\n",
                count_games, j, h_position_history[j], game_ply, h_position);
            }
          }

          /* Ecriture du coup des noirs : */
          hash_ply++;
          fprintf (hash_test_file, "%d. ", hash_ply);
          fprintf (hash_test_file, lan_move);
          fprintf (hash_test_file, " %8x", h_position);
          fprintf (hash_test_file, " ");

        }
      }
      fprintf (hash_test_file, "\n");
      uci.ucinewgame ();
    }

    fclose (games_file);
    fclose (hash_test_file);

    tools.send_output (true, "   %s write %d games in the hash test file.\n\n",
      PROGRAM_NAME, count_games, HASH_TEST_FILE);
  }
}
/* End of function : test_hash */

/*******************************************************************************
  Function : trace

  Description : Pour tracer l'appel  chaque fonction du programme.

  Validation level : 1 - debug.trace
*******************************************************************************/
void trace (char fn_name[], bool fn_print, int *fn_calls) {

#ifdef DEBUG
  total_calls++;

#ifdef DEBUG_HIGH
  if (fn_print) {

/*
    debug.print ("%9d (%9d) - %20s", total_calls, nodes, fn_name);
*/
    debug.print ("(%9d) - %20s", nodes, fn_name);
  }
#endif /* DEBUG_HIGH */

  (*fn_calls)++;
#endif /* DEBUG */
}
/* End of function : trace */

/*******************************************************************************
  Object declaration :
*******************************************************************************/

debug_module_s debug = {

  close_debug_files,
  dump,
  new_debug_files,
  open_debug_files,
  print,
  print_informations,
  print_statistics,
  test_hash,
  trace
};
