/*******************************************************************************
# Copyright (c) 2009 <copyright Christophe J. Mandin>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
/*******************************************************************************
  File : sissa.c

  Author  : Christophe J. Mandin
  Date    : 2009/01/16
  Version : 0.12.0

  Description : Sissa est un moteur d'checs.
*******************************************************************************/

/*******************************************************************************
  Include files :
*******************************************************************************/

#include "book.h"
#include "data.h"
#include "debug.h"
#include "evaluate.h"
#include "hash.h"
#include "moves.h"
#include "search.h"
#include "sissa.h"
#include "tools.h"
#include "uci.h"

/*******************************************************************************
  Data :
*******************************************************************************/

#ifdef DEBUG
int computer_to_play_calls = 0;
int init_game_calls = 0;
int main_calls = 0;
#endif /* DEBUG */

unsigned int total_nodes = 0;

int computer_game_stage = UNDEFINED_GS;

/******************************************************************************
  Internal functions prototypes :
******************************************************************************/

void computer_to_play (void);

/*******************************************************************************
  External functions :
*******************************************************************************/

/*******************************************************************************
  Function : init_game

  Description : Met  jour les structures de donnes  partir de la position
                donne en entre.

  Validation level : 1 - sissa.init_game
*******************************************************************************/
void init_game (const char string[]) {

/* ------------------------------- Local data ------------------------------- */

char c;
char *pointer;
char position[STRING_LENGTH];

int color;
int i, j;
int piece;
int x, y;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("init_game\n", true, &init_game_calls);
#endif /* DEBUG */

  game_ply = 0;
  piece_count = 0;

  memset (board, 0, sizeof (board));

  strcpy (position, string);

  /* Piece placement data : */
  pointer = strtok (position, " ");
  i = 0;
  c = pointer[i];

  for (y = 8; y >= 1; y--) {

    for (x = A; x <= H; ) {

      if (c >= '1' && c <= '8') {

        for (j = 0; j < c - '0'; j++) {

          board[x][y].piece = NONE;
          x++;
        }
      }
      else {

        switch (c) {
          case 'b': piece = BISHOP; color = BLACK; piece_count++; break;
          case 'k': piece = KING;   color = BLACK;
            king[BLACK].x = x;
            king[BLACK].y = y;
            break;
          case 'n': piece = KNIGHT; color = BLACK; piece_count++; break;
          case 'p': piece = PAWN;   color = BLACK; break;
          case 'q': piece = QUEEN;  color = BLACK; piece_count++; break;
          case 'r': piece = ROOK;   color = BLACK; piece_count++; break;

          case 'B': piece = BISHOP; color = WHITE; piece_count++; break;
          case 'K': piece = KING;   color = WHITE;
            king[WHITE].x = x;
            king[WHITE].y = y;
            break;

          case 'N': piece = KNIGHT; color = WHITE; piece_count++; break;
          case 'P': piece = PAWN;   color = WHITE; break;
          case 'Q': piece = QUEEN;  color = WHITE; piece_count++; break;
          case 'R': piece = ROOK;   color = WHITE; piece_count++; break;
        }

        board[x][y].piece = piece;
        board[x][y].color = color;

        x++;
      }
      c = pointer[++i];
    }
    if (y > 1) {

      if (c != '/') {

        tools.send_output (false, "ERROR : Bad position (rank)\n");
      }

      c = pointer[++i];
    }
  }

  tools.update_pieces ();

  /* Active color : */
  pointer = strtok (NULL, " ");
  if (pointer[0] == 'w') {

    active_color = WHITE;
  }
  else if (pointer[0] == 'b') {

    active_color = BLACK;
  }
  else {

    tools.send_output (false, "ERROR : Bad position (unknown color)\n");
  }

  /* Castling availability : */
  pointer = strtok (NULL, " ");
  i = 0;
  c = pointer[i];

  if (c == '-') {
  }
  else {
    if (c == 'K') {

      CASTLE_STATUS |= 0x1;
      c = pointer[++i];
    }

   if (c == 'Q') {

      CASTLE_STATUS |= 0x2;
      c = pointer[++i];
    }

    if (c == 'k') {

      CASTLE_STATUS |= 0x4;
      c = pointer[++i];
    }

    if (c == 'q') {

      CASTLE_STATUS |= 0x8;
    }
  }

  /* En-passant target square : */
  pointer = strtok (NULL, " ");
  i = 0;
  c = pointer[i];

  if (c == '-') {

    EP_SQUARE.x = 0;
    EP_SQUARE.y = 0;
  }
  else {

    if (c < 'a' || c > 'h') {

      tools.send_output (false, "ERROR : Bad position (file for ep square)");
    }

    y = c - 'a';
    c = pointer[++i];

    x = c - '1';

    EP_SQUARE.x = x;
    EP_SQUARE.y = y;
  }

  /* Halfmove clock :
     The fifth field is a nonnegative integer representing the halfmove clock.
     This number is the count of halfmoves (or ply) since the last pawn advance
     or capturing move.  This value is used for the fifty move draw rule. */
  pointer = strtok (NULL, " ");
  c = pointer[0];

  if (!isdigit(c)) {

    tools.send_output (false,
      "ERROR : Bad position (Halfmove is not a digit)");
  }
  else {

    halfmove = atoi (&pointer[0]);
  }

  /* Fullmove number :
     The sixth and last field is a positive integer that gives the fullmove
     number. This will have the value "1" for the first move of a game for
     both White and Black. It is incremented by one immediately after each
     move by Black. */
  pointer = strtok (NULL, " ");
  c = pointer[0];

  if (!isdigit(c)) {

    tools.send_output (false,
      "ERROR : Bad position (Fullmove is not a digit)");
  }
  else {

    fullmove = atoi (&pointer[0]);
  }

  computer_color = active_color^1;

  result = NO_RESULT;

  /* Calcule la valeur de hachage de la position initiale : */
  h_position = hash.compute_h_position ();
  h_position_history[0] = h_position;
}
/* End of function : init_game */

/*******************************************************************************
  Object declaration :
*******************************************************************************/

sissa_module_s sissa = {

  init_game
};

/*******************************************************************************
  Internal functions :
*******************************************************************************/

/******************************************************************************
  Function : computer_to_play

  Description : Lancement de la rflexion du programme et traitement du coup
                choisi, ainsi que du rsultat ventuel de la partie.

  Validation level : 1 - sissa.computer_to_play
******************************************************************************/
void computer_to_play (void) {

/* ------------------------------- Local data ------------------------------- */

char move_string[STRING_LENGTH];
char output[STRING_LENGTH];

int elapsed_time;
int game_stage;
int i;

move_s computer_move;

/* ------------------------------ Function code ----------------------------- */

#ifdef DEBUG
debug.trace ("computer_to_play\n", true, &computer_to_play_calls);
#endif /* DEBUG */

  game_stage = evaluate.check_game_stage (active_color);

  if (computer_game_stage != game_stage) {

    if (uci_mode) {

      printf ("info string %s is in the %s.\n", PROGRAM_NAME,
        game_stage == ENDGAME ? "end of game" :
        game_stage == MIDDLE ? "middle of game" : "opening");
    }
    else {

      tools.send_output (true, "   %s is in the %s.\n\n", PROGRAM_NAME,
        game_stage == ENDGAME ? "end of game" :
        game_stage == MIDDLE ? "middle of game" : "opening");
    }

    computer_game_stage = game_stage;
  }

  start_time = time (0);
  cpu_start = clock ();
  search.think (&computer_move);
  cpu_end = clock ();

  total_nodes += nodes;

  /* Si la partie n'est pas termine : */
  if (   (   (computer_color == BLACK && result != BLACK_IS_MATED)
          || (computer_color == WHITE && result != WHITE_IS_MATED))
      && result != DRAW_FIFTY_MOVES
      && result != DRAW_REPETITION
      && result != DRAW_STALEMATE) {

    /* alors le coup est jou sur l'chiquier : */
    tools.convert_move (computer_move, output);
    moves.make (computer_move);

    /* Teste si la partie est nulle aprs avoir jou le dernier coup : */
    if (tools.is_repeated ()) {

      result = DRAW_REPETITION;
    }
    else if (halfmove > 100) {

      result = DRAW_FIFTY_MOVES;
    }

    tools.update_pieces ();

    if (uci_mode) {

      elapsed_time = (int) difftime (time (0), start_time);

      tools.send_output (true, "bestmove %s\n", output);
    }
    else {

      tools.send_output (true, "\n   %s plays %s\n", PROGRAM_NAME, output);

      if (nodes) {

        tools.send_output (true, "\n   Elapsed time %.1f\n",
          (cpu_end - cpu_start) / (double) CLOCKS_PER_SEC);
        tools.send_output (true, "   Nodes        %d (%d nps)\n", nodes,
          (int) ((float) nodes / ((cpu_end - cpu_start) /
          (double) CLOCKS_PER_SEC)));
        tools.send_output (true, "   Total nodes  %d\n",
          (int) (total_nodes / 1000000.0));
      }
    }
  }

  /* Affichage d'un ventuel rsultat de la partie : */
  if (result) {

    if (uci_mode) {

      printf ("info string ");
    }
    else {

      tools.send_output (true, "\n   ");
    }

    switch (result) {
      case WHITE_IS_MATED :
        tools.send_output (true, "0-1 (black mates)\n");
        break;
      case BLACK_IS_MATED :
        tools.send_output (true, "1-0 (white mates)\n");
        break;
      case DRAW_FIFTY_MOVES :
        tools.send_output (true, "1/2-1/2 (draw - fifty move)\n");
        break;
      case DRAW_REPETITION :
        tools.send_output (true, "1/2-1/2 (draw - 3 repetitions)\n");
        break;
      case DRAW_STALEMATE :
        tools.send_output (true, "1/2-1/2 (draw - stalemate)\n");
        break;
      default :
        tools.send_output (true, "Warning unknown result (%d)\n", result);
        break;
    }
  }
}
/* End of function : computer_to_play */

/*******************************************************************************
  Function : main

  Description : Point d'entre du programme.

  Validation level : 1 - sissa.main
*******************************************************************************/
int main (int argc, char *argv[]) {

/* ------------------------------- Local data ------------------------------- */

char input[STRING_LENGTH];
char lan_move[6];
char search_type_char;

int i;
int moves_number;
int search_type;

move_s move;
move_s legal_moves[MOVES_LIMIT];

/* ------------------------------ Function code ----------------------------- */

  /* Pas d'utilisation de tampons sur les flux d'entre et de sortie : */
  setbuf (stdout, NULL);
  setbuf (stdin, NULL);

  /* Pas d'interruption  partir du clavier, le programme se terminera par la
     commande "quit" : */
  signal (SIGINT, SIG_IGN);

#ifdef DEBUG
  /* Les changes entre le programme et l'extrieur sont mmoriss dans un
     fichier journal : */
  if ((log_file = fopen (LOG_FILE, "w")) == NULL) {

    printf ("ERROR : %s can't open the log file %s\n",
            PROGRAM_NAME, LOG_FILE);
  }

  /* Ouverture des fichiers utiliss pour dboguer le programme : */
  debug.open_debug_files ();
#endif /* DEBUG */

#ifdef DEBUG
debug.trace ("main\n", true, &main_calls);
#endif /* DEBUG */

  /* Prsentation du programme en mode console : */
  tools.send_output (true, "\n       %s by %s\n",
                     PROGRAM_NAME, PROGRAM_AUTHOR);
  tools.send_output (true, "           Version %s - %s\n\n",
                     PROGRAM_VERSION, PROGRAM_YEAR);
  tools.send_output (true, "        Type help to display help !\n");

  /* Par dfaut, le programme dmarre en mode console : */
  use_book    = true;
  forced_mode = false;
  show_board  = true;
  uci_mode    = false;

  /* Initialisation du coup sans signification : */
  memset (&null_move, 0, sizeof (null_move));

  /* Initialisation d'une nouvelle squence de nombres pseudo-alatoires : */
  srand(64);

  /* Initialisation du module de hachage : */
  hash.init_h_values ();

  /* Initialise la partie avec la position de dpart classique : */
  sissa.init_game (classic_start_position);

  /* Boucle infine en attente d'une entre : */
  while (true) {

    /* Si le programme a le trait : */
    if (computer_color == active_color && !forced_mode && !result) {

      /* alors la rflexion du programme est lance : */
      computer_to_play ();

      if (uci_mode) {

        forced_mode = true;
      }
    }

    /* En mode console, affichage de l'chiquier : */
    if (!uci_mode) {

      if (   show_board
          && strcmp (input, "book") && strcmp (input, "nobook")
          && strcmp (input, "dump")
          && strcmp (input, "eval")
          && strcmp (input, "force")
          && strcmp (input, "game")
          && strcmp (input, "hash")
          && strcmp (input, "hashtest")
          && strcmp (input, "help")
          && strcmp (input, "hpos")
          && strcmp (input, "makebook")
          && strncmp (input, "moves", 5)
          && strcmp (input, "select")
          && strcmp (input, "thinking")) {

        if (forced_mode) {
          tools.send_output (true, "   Evaluation   %5.2f\n\n",
            (float) evaluate.evaluate_position (true) / 100);
        }
        else {

          printf ("\n");
        }

        tools.print_board (stdout);
      }

      printf ("%s -> ", PROGRAM_NAME);
    }

    /* Lecture d'une entre : */
    tools.read_input (input);

    if (!uci_mode) {

      printf ("\n");
    }

    /* Vrifie et convertit le coup entr au format LAN (Long Algebraic
       Notation) : */
    if (   isalpha ((int) input[0])
        && isdigit ((int) input[1])
        && isalpha ((int) input[2])
        && isdigit ((int) input[3])) {

      if (tools.convert_lan_move (input, &move)) {

        if (tools.verify_move (input)) {

          moves.make (move);
          tools.update_pieces ();
        }
        else {

          tools.send_output (true, "   ERROR : The move is not legal !\n");

          if (forced_mode) {

            tools.send_output (true, "\n");
          }
        }
      }
      else {

        tools.send_output (true,
          "   ERROR : This is not a possible move !\n");

        if (forced_mode) {

          tools.send_output (true, "\n");
        }
      }
    }
    else {

      /* Commandes UCI : */

      if (!strncmp (input, "debug", 5)) {

        uci.debug_cmd ();
      }
      else if (!strncmp (input, "go", 2)) {

        if (uci_mode) {

          uci.go (input);
        }

        forced_mode = false;

        computer_color = active_color;
      }
      else if (!strcmp (input, "isready")) {

        uci.isready ();
      }
      else if (!strcmp (input, "ponderhit")) {

        uci.ponderhit ();
      }
      else if (!strncmp (input, "position", 8)) {

        uci.position (input);
      }
      if (!strcmp (input, "quit") || !strcmp (input, "exit")) {

        uci.quit (EXIT_SUCCESS);
      }
      else if (!strncmp (input, "register", 8)) {

        uci.registration ();
      }
      else if (!strncmp (input, "setoption", 9)) {

        uci.setoption ();
      }
      else if (!strcmp (input, "stop")) {

        uci.stop ();
      }
      else if (!strcmp (input, "uci")) {

        uci.uci_cmd ();
      }
      else if (!strcmp (input, "ucinewgame") || !strcmp(input, "new")) {

        uci.ucinewgame ();
      }

      /* Autres commandes : */

      else if (!strcmp (input, "book")) {
        use_book = true;
        printf ("   %s will use book moves.\n\n", PROGRAM_NAME);
      }
      else if (!strcmp (input, "nobook")) {
        use_book = false;
        printf ("   %s will not use book moves.\n\n", PROGRAM_NAME);
      }
      else if (!strcmp (input, "dump")) {

        debug.dump ();
      }
      else if (!strcmp (input, "eval")) {

        tools.send_output (true, "   Evaluation : %5.2f\n\n",
          (float) evaluate.evaluate_position (true) / 100);
      }
      else if (!strcmp (input, "force")) {

        forced_mode ^= 1;
        computer_color = active_color ^1;

        if (forced_mode) {

          tools.send_output (true, "   %s is in forced mode.\n\n",
            PROGRAM_NAME);
        }
        else {

          tools.send_output (true, "   %s is not in forced mode.\n\n",
            PROGRAM_NAME);
        }
      }
      else if (!strcmp (input, "game")) {

        if (game_ply != 0) {

          for (i = 0; i < game_ply; i++) {

            if (played_moves[i].color == WHITE) {

              tools.send_output (true, "   %2d) ", i / 2 + 1);
            }

            if (played_moves[i].color == BLACK ) {

              tools.send_output (true, " - ");
            }

            tools.send_output (true, "%c %c%d%c%c%d%c",
              PIECE(played_moves[i].piece),
              played_moves[i].from.x + (int) 'a' - 1,
              played_moves[i].from.y,
              played_moves[i].captured_form ? 'x' : '-',
              played_moves[i].to.x + (int) 'a' - 1,
              played_moves[i].to.y,
              played_moves[i].status == PROMOTION ?
                played_moves[i].promotion + (int) 'a' - 1 : ' ');

            if (played_moves[i].color == BLACK) {

              tools.send_output (true, "\n");
            }
          }

          if (played_moves[i - 1].color == WHITE) {

            tools.send_output (true, "\n");
          }

          tools.send_output (true, "\n");
        }
        else {

          tools.send_output (true, "   WARNING : The game list is empty.\n\n");
        }
      }
      else if (!strcmp (input, "help")) {

        printf ("*******************************************************\n\n");
        printf (" UCI commands :\n\n");
        printf ("  %s accepts all UCI commands : debug, go,\n", PROGRAM_NAME);
        printf ("  isready, ponderhit, position, quit, register,\n");
        printf ("  setoption,stop, uci, ucinewgame.\n\n");
        printf (" Other commands :\n\n");
        printf ("  book/nobook  : To use (or not) the openings book\n");
        printf ("  dump         : To dump the data\n");
        printf ("  eval         : To obtain a static evaluation of the\n");
        printf ("                 position.\n");
        printf ("  exit or quit : To quit the program\n");
        printf ("  force        : To toggle the forced mode (on/off)\n");
        printf ("  game         : List all moves of the game\n");
        printf ("  help         : To get help about the program\n");
        printf ("  makebook     : To build openings book\n");
        printf ("  moves [c/p]  : To display legal moves (default)\n");
        printf ("    (c for captures only - p for pseudo moves)\n");
        printf ("  selectmove   : To select a move in the openings book\n");
        printf (" For more informations, read the readme.txt file.\n\n");
        printf ("*******************************************************\n\n");
      }
      else if (!strncmp (input, "moves", 5)) {

        search_type = LEGAL_MOVES;

        if (strlen (input) == 7) {

          sscanf (input + 6, "%c", &search_type_char);

          if (search_type_char == 'c') {

            search_type = CAPTURES_ONLY;
          }
          else if (search_type_char == 'p') {

            search_type = PSEUDO_MOVES;
          }
          else {

             tools.send_output (true,
               "   WARNING : Search type must be 'c' or 'p'\n\n");
          }
        }

        moves_number = moves.generate (legal_moves, search_type);

        for (i = 1; i <= moves_number; i++) {

          tools.send_output (true, "   %c%d%c%c%d%c ",
            legal_moves[i].from.x + (int) 'a' - 1, legal_moves[i].from.y,
            legal_moves[i].captured_form ? 'x' : '-',
            legal_moves[i].to.x + (int) 'a' - 1, legal_moves[i].to.y,
            legal_moves[i].status == PROMOTION ?
              legal_moves[i].promotion + (int) 'a' - 1 : ' ');

          moves.make (legal_moves[i]);

          tools.send_output (true, "%5.2f\n", active_color ?
            (float) evaluate.evaluate_position (false) / 100 :
            - (float) evaluate.evaluate_position (false) / 100);

          moves.unmake (legal_moves[i]);
        }

        if (moves_number != 0) {

          tools.send_output (true, "\n");
        }

        tools.send_output (true, "   Number of legal moves : %d\n",
          moves_number);

        printf("\n");
      }
      else if (!strcmp (input, "hash")) {

      int i = 0;

        tools.send_output (true, "   Historique des valeurs de hachage des "
          "positions :\n\n");
        for (i = 0; i <= game_ply; i++) {
          tools.send_output (true, "   h_position_history[%2d] %x\n",
            i, h_position_history[i]);
        }
        tools.send_output (true, "\n");
      }
      else if (!strcmp (input, "hashtest")) {

        debug.test_hash ();
      }
      else if (!strcmp (input, "hpos")) {

      h_value_s current_h_position;

        current_h_position = hash.compute_h_position ();
        tools.send_output (true, "   h_position %x\n\n", current_h_position);
      }
      else if (!strcmp (input, "makebook")) {

        book.make_book ();
      }
      else if (!strcmp (input, "select")) {

        book.select_move (&move);
      }
      /* Tests de performance du programme : */
      else if (!strcmp (input, "test")) {

      int x, y, i, j;
      unsigned int nb_tests;
      move_s legal_moves[MOVES_LIMIT];
      int moves_number;

        nb_tests = 1000000;

        cpu_start = clock ();
        for (i = 0; i < nb_tests; i++) {

          /* Tests sur la gnration des coups et le dplacement des pices : */
          moves_number = moves.generate (legal_moves, LEGAL_MOVES);
          do {
            moves.make (legal_moves[moves_number]);
            moves.unmake (legal_moves[moves_number]);

          } while (--moves_number);
        }
        cpu_end = clock ();

        tools.send_output (true, "\n   Elapsed time for test : %.1f\n",
          (cpu_end - cpu_start) / (double) CLOCKS_PER_SEC);
      }
      else if (!uci_mode && strcmp (input, "go")) {

        tools.send_output (true, "   Unknown input : %s\n", input);

        if (forced_mode) {

          tools.send_output (true, "\n");
        }
      }
    }
  }
}
/* End of function : main */
